"""
# -*- coding: utf-8 -*-

__all__ = ["__version__"]

#from . import blah, blah, blah
# from .blah import * or something specific

__author__ = "Christopher Lam"
__email__ = "c.lam@ufl.com"
__license__ = "MIT License"
__description__ = "Fisher Information-based radial velocity observation scheduling"
"""